from typing import Any

import pandas as pd
import xarray as xr

from .utils import _load_JSON


def open(dataset: str) -> Any:
    """Opens a dataset.

    Parameters
    ----------
    dataset : str
        One of "sentinel" or "spectral". The sentinel dataset is loaded as a
        :code:`xarray.DataArray` with a sample image of the Sentinel-2 satellite
        (10 m bands). The spectral dataset is loaded as a :code:`pandas.DataFrame`
        with Landsat 8 reflectance samples of three different land covers.

    Returns
    -------
    Any
        Loaded dataset.

    Examples
    --------
    Open the :code:`sentinel` dataset:

    >>> import spyndex
    >>> spyndex.datasets.open("sentinel")
    <xarray.DataArray (band: 4, x: 300, y: 300)>
    Coordinates:
    * band     (band) <U3 'B02' 'B03' 'B04' 'B08'
    Dimensions without coordinates: x, y

    Open the :code:`spectral` dataset:

    >>> spt = spyndex.datasets.open("spectral")
    >>> spt.dtypes
    SR_B1     float64
    SR_B2     float64
    SR_B3     float64
    SR_B4     float64
    SR_B5     float64
    SR_B6     float64
    SR_B7     float64
    ST_B10    float64
    class      object
    dtype: object
    >>> spt.shape
    (120, 9)
    """

    datasets = {"sentinel": "S2_10m.json", "spectral": "spectral.json"}

    if dataset not in list(datasets.keys()):
        raise Exception(
            f"{dataset} is not a valid dataset. Please use one of ['sentinel','spectral']"
        )

    ds = _load_JSON(datasets[dataset])

    if dataset == "sentinel":
        ds = xr.DataArray(
            ds, dims=("band", "x", "y"), coords={"band": ["B02", "B03", "B04", "B08"]}
        )
    elif dataset == "spectral":
        ds = pd.DataFrame(ds)

    return ds



import re
from typing import Any, List, Optional, Union

import dask
import dask.array as da
import dask.dataframe as dd
import numpy as np
import pandas as pd
import xarray as xr

from .utils import _check_params, _get_indices, _has_ee_image, _has_ee_number, _maybe_import_earthengine


def computeIndex(
    index: Union[str, List[str]],
    params: Optional[dict] = None,
    online: bool = False,
    returnOrigin: bool = True,
    coordinate: str = "index",
    **kwargs,
) -> Any:
    """Computes one or more Spectral Indices from the Awesome Spectral Indices list.

    Parameters
    ----------
    index : str | list[str]
        Index or list of indices to compute. Check all available indices from the
        `Awesome Spectral Indices Repository <https://github.com/davemlz/awesome-spectral-indices>`_.
    params: dict, default = None
        Parameters used as inputs for the computation. The input data must be compatible
        with Overloaded Operators. Some inputs' types supported are pandas series,
        numpy arrays, xarray objects and numeric objects. Earth Engine objects are also
        compatible when using eemont.
    online : bool, default = False
        Whether to retrieve the most recent list of indices directly from the GitHub
        repository and not from the local copy.
    returnOrigin : bool, default = True
        Whether to return multiple indices as an object with the same type as the inputs.
            - :code:`pandas.Series`: Returns a :code:`pandas.DataFrame`.
            - :code:`numpy.ndarray`: Returns a :code:`numpy.ndarray`.
            - :code:`xarray.DataArray`: Returns a :code:`xarray.DataArray`.
            - :code:`ee.Image`: Returns a :code:`ee.Image`.
            - :code:`ee.Number`: Returns a :code:`ee.List`.
            - :code:`dask.Array`: Returns a :code:`dask.Array`.
            - :code:`dask.Series`: Returns a :code:`dask.DataFrame`.
        When numeric objects are used in combination with other objects, the type of the
        other object is returned. If false, a list is returned.
    coordinate : str, default = "index"
        Name of the coordinate used to concatenate :code:`xarray.DataArray` objects when
        :code:`returnOrigin = True`.
    kwargs:
        Parameters used as inputs for the computation as keyword pairs. Ignored when
        params is defined.

        .. versionadded:: 0.0.5

    Returns
    -------
    Any
        Computed Spectral Indices according to the inputs' type.

    See Also
    --------
    computeKernel : Computes a kernel :code:`k(a,b)`.

    Examples
    --------
    Compute a Spectral Index by passing the required :code:`params` dictionary:

    >>> import spyndex
    >>> spyndex.computeIndex(
    ...     index = "NDVI",
    ...     params = {
    ...         "N": 0.643,
    ...         "R": 0.175
    ...     }
    ... )
    0.5721271393643031

    Compute a Spectral Index by passing the required :code:`params` as keyword pairs:

    >>> spyndex.computeIndex("NDVI",N = 0.643,R = 0.175)
    0.5721271393643031

    Two or more Spectral Indices can also be computed:

    >>> spyndex.computeIndex(
    ...     index = ["NDVI","SAVI"],
    ...     params = {
    ...         "N": 0.643,
    ...         "R": 0.175,
    ...         "L": 0.5
    ...     }
    ... )
    [0.5721271393643031, 0.5326251896813354]

    Spyndex is versatile! Let's compute Spectral Indices from a :code:`numpy.ndarray`:

    >>> import numpy as np
    >>> R = np.random.normal(0.12,0.05,10000)
    >>> G = np.random.normal(0.34,0.07,10000)
    >>> N = np.random.normal(0.67,0.12,10000)
    >>> spyndex.computeIndex(
    ...     index = ["NDVI","SAVI","GNDVI"],
    ...     params = {
    ...         "N": N,
    ...         "R": R,
    ...         "G": G,
    ...         "L": 0.5
    ...     }
    ... )
    array([[0.57190873, 0.63776266, 0.52554653, ..., 0.692647  , 0.72013087,
            0.57576994],
           [0.5494994 , 0.60604837, 0.47157809, ..., 0.60647869, 0.65887439,
            0.52585032],
           [0.33304486, 0.46408771, 0.28007567, ..., 0.35734698, 0.28536337,
            0.50212151]])

    Now, let's try a :code:`pandas.DataFrame`:

    >>> import pandas as pd
    >>> df = pd.DataFrame({"Red":R,"Green":G,"NIR":N})
    >>> spyndex.computeIndex(
    ...     index = ["NDVI","SAVI","GNDVI"],
    ...     params = {
    ...         "N": df["NIR"],
    ...         "R": df["Red"],
    ...         "G": df["Green"],
    ...         "L": 0.5
    ...     }
    ... )
            NDVI      SAVI     GNDVI
    0     0.571909  0.549499  0.333045
    1     0.637763  0.606048  0.464088
    2     0.525547  0.471578  0.280076
    3     0.498328  0.443842  0.514775
    4     0.625445  0.512757  0.227829
    ...        ...       ...       ...
    9995  0.706123  0.604131  0.233519
    9996  0.731205  0.630090  0.389462
    9997  0.692647  0.606479  0.357347
    9998  0.720131  0.658874  0.285363
    9999  0.575770  0.525850  0.502122

    What about a :code:`xarray.DataArray`?

    >>> import xarray as xr
    >>> da = xr.DataArray(np.array([G,R,N]).reshape(3,100,100),
    ...     dims = ("band","x","y"),
    ...     coords = {"band": ["Green","Red","NIR"]})
    >>> spyndex.computeIndex(
    ...     index = ["NDVI","SAVI","GNDVI"],
    ...     params = {
    ...         "N": da.sel(band = "NIR"),
    ...         "R": da.sel(band = "Red"),
    ...         "G": da.sel(band = "Green"),
    ...         "L": 0.5
    ...     }
    ... )
    <xarray.DataArray (index: 3, x: 100, y: 100)>
    Coordinates:
        * index    (index) <U5 'NDVI' 'SAVI' 'GNDVI'
    Dimensions without coordinates: x, y

    Now let's try :code:`dask`!

    >>> import dask.array as da
    >>> array_shape = (10000,10000)
    >>> chunk_size = (1000,1000)
    >>> dask_array = da.array([
    ...     da.random.normal(0.6,0.10,array_shape,chunks = chunk_size),
    ...     da.random.normal(0.1,0.05,array_shape,chunks = chunk_size),
    ...     da.random.normal(0.3,0.02,array_shape,chunks = chunk_size)
    ... ])
    >>> spyndex.computeIndex(
    ...     index = ["NDVI","SAVI","GNDVI"],
    ...     params = {
    ...         "N": dask_array[0],
    ...         "R": dask_array[1],
    ...         "G": dask_array[2],
    ...         "L": 0.5
    ...     }
    ... ).compute()

    And a :code:`dask.DataFrame`?

    >>> import dask.dataframe as dd
    >>> df = pd.DataFrame({
    ...     "NIR": np.random.normal(0.6,0.10,1000),
    ...     "RED": np.random.normal(0.1,0.05,1000),
    ...     "GREEN": np.random.normal(0.3,0.02,1000),
    ... })
    >>> df = dd.from_pandas(df,npartitions = 10)
    >>> spyndex.computeIndex(
    ...     index = ["NDVI","SAVI","GNDVI"],
    ...     params = {
    ...         "N": df["NIR"],
    ...         "R": df["RED"],
    ...         "G": df["GREEN"],
    ...         "L": 0.5
    ...     }
    ... ).compute()
    """

    if params is None:
        params = kwargs

    if not isinstance(index, list):
        index = [index]

    indices = _get_indices(online)
    names = list(indices.keys())

    _maybe_import_earthengine(params)

    result = []
    for idx in index:
        if idx not in names:
            raise Exception(f"{idx} is not a valid Spectral Index!")
        else:
            _check_params(idx, params, indices)
            result.append(eval(indices[idx]["formula"], {}, params))

    if len(result) == 1:
        result = result[0]
    else:
        if returnOrigin:
            if isinstance(result[0], np.ndarray):
                result = np.array(result)
            elif isinstance(result[0], pd.Series):
                result = pd.DataFrame(dict(zip(index, result)))
            elif isinstance(result[0], xr.DataArray):
                result = [x.reset_coords(drop=True) for x in result]
                result = xr.concat(result, dim=coordinate).assign_coords(
                    {coordinate: index}
                )
            elif _has_ee_image(params):
                import ee, eemont
                result = ee.Image(result).rename(index)
            elif _has_ee_number(params):
                import ee, eemont
                result = ee.List(result)
            elif isinstance(result[0], dask.array.Array):
                result = da.array(result)
            elif isinstance(result[0], dask.dataframe.Series):
                result = dd.concat(result, axis="columns")
                result.columns = index

    return result


def computeKernel(kernel: str, params: Optional[dict] = None, **kwargs) -> Any:
    """Computes a kernel :code:`k(a,b)`.

    Kernel parameters are used for kernel indices like the kNDVI that requires the
    :code:`kNN` (:code:`k(N,N)`) and :code:`kNR` (:code:`k(N,R)`) parameters.

    Parameters
    ----------
    kernel : str
        Kernel to use. One of 'linear', 'poly' or 'RBF'.
    params : dict
        Parameters to use for the kernel computation.
        For :code:`kernel = 'linear'`, the parameters 'a' (band A) and 'b' (band B) must
        be declared. For :code:`kernel = 'RBF'`, the parameters 'a' (band A), 'b' (band B)
        and 'sigma' (length-scale) must be declared. For :code:`kernel = 'poly'`, the
        parameters 'a' (band A), 'b' (band B), 'p' (kernel degree) and 'c' (trade-off)
        must be declared.
    kwargs:
        Parameters used as inputs for the computation as keyword pairs. Ignored when
        params is defined.

        .. versionadded:: 0.0.5

    Returns
    -------
    Any
        Computed kernel.

    See Also
    --------
    computeIndex : Computes one or more Spectral Indices from the Awesome Spectral Indices
        list.

    Examples
    --------
    Compute a kernel index with the help of :code:`spyndex.computeKernel()`:

    >>> import spyndex
    >>> spyndex.computeIndex(
    ...     index = "kNDVI",
    ...     params = {
    ...         "kNN": 1.0,
    ...         "kNR": spyndex.computeKernel(
    ...             kernel = "RBF",
    ...             params = {
    ...                 "a" : 0.68, "b": 0.13, "sigma": (0.68 + 0.13) / 2
    ...             }
    ...         )
    ...     }
    ... )
    0.4309459271768674

    Parameters can also be passed as keyword pairs:

    >>> import spyndex
    >>> spyndex.computeIndex(
    ...     index = "kNDVI",
    ...     kNN = 1.0,
    ...     kNR = spyndex.computeKernel("RBF",a = 0.68,b = 0.13,sigma = (0.68 + 0.13) / 2)
    ... )
    0.4309459271768674

    Use the polynomial kernel:

    >>> import spyndex
    >>> spyndex.computeIndex(
    ...     index = "kNDVI",
    ...     params = {
    ...         "kNN": spyndex.computeKernel(
    ...             kernel = "poly",
    ...             params = {
    ...                 "a" : 0.68,
    ...                 "b": 0.68,
    ...                 "p": 2.0,
    ...                 "c": spyndex.constants.c.default
    ...             }
    ...         ),
    ...         "kNR": spyndex.computeKernel(
    ...             kernel = "poly",
    ...             params = {
    ...                 "a" : 0.68,
    ...                 "b": 0.13,
    ...                 "p": 2.0,
    ...                 "c": spyndex.constants.c.default
    ...             }
    ...         )
    ...     }
    ... )
    0.2870700138954041

    Now let's try a :code:`numpy.ndarray`:

    >>> import numpy as np
    >>> R = np.random.normal(0.12,0.05,10000)
    >>> N = np.random.normal(0.67,0.12,10000)
    >>> spyndex.computeIndex(
    ...     index = "kNDVI",
    ...     params = {
    ...         "kNN": 1.0,
    ...         "kNR": spyndex.computeKernel(
    ...             kernel = "RBF",
    ...             params = {
    ...                 "a" : N,
    ...                 "b" : R,
    ...                 "sigma" : np.mean([N,R],0)
    ...            }
    ...         )
    ...     }
    ... )
    array([0.36776416, 0.57727362, 0.5252302 , ..., 0.5209451 , 0.53162097,
       0.67689597])

    It's time for a :code:`pandas.DataFrame`!

    >>> import pandas as pd
    >>> R = np.random.normal(0.12,0.05,10000)
    >>> N = np.random.normal(0.67,0.12,10000)
    >>> df = pd.DataFrame({"Red":R,"NIR":N})
    >>> spyndex.computeIndex(
    ...     index = "kNDVI",
    ...     params = {
    ...         "kNN": 1.0,
    ...         "kNR": spyndex.computeKernel(
    ...             kernel = "RBF",
    ...             params = {
    ...                 "a" : df["NIR"],
    ...                 "b" : df["Red"],
    ...                 "sigma" : df.mean(1)
    ...            }
    ...         )
    ...     }
    ... )
    0       0.468294
    1       0.535752
    2       0.745249
    3       0.402761
    4       0.432528
            ...
    9995    0.475168
    9996    0.482034
    9997    0.403363
    9998    0.489537
    9999    0.508163
    Length: 10000, dtype: float64
    """

    if params is None:
        params = kwargs

    kernels = {
        "linear": "a * b",
        "poly": "((a * b) + c) ** p",
    }

    _maybe_import_earthengine(params)

    if _has_ee_image(params) or _has_ee_number(params):
        import ee, eemont
        kernels["RBF"] = "exp((-1.0 * (a - b) ** 2.0)/(2.0 * sigma ** 2.0))"
        result = params["a"].expression(kernels[kernel], params)
    else:
        kernels["RBF"] = "np.exp((-1.0 * (a - b) ** 2.0)/(2.0 * sigma ** 2.0))"
        params["np"] = np
        result = eval(kernels[kernel], params)

    return result



from typing import Optional, Union

import numpy as np
import pandas as pd
import seaborn as sns

import spyndex

from .utils import _check_params, _get_indices


def heatmap(index: str, x: str, y: str, params: Optional[dict] = None, online: bool = False, **kwargs):
    """Plot all posible index values as a color-encoded matrix.

    Parameters
    ----------
    index : str
        Index to plot.
    x : str
        Band to plot in the x-axis.
    y : str
        Band to plot in the y-axis.
    params: dict
        Parameters for all remaining bands that are not used in :code:`x` nor :code:`y`.
        This dictionary must store just float values.
    online : bool, default = False
        Whether to retrieve the most recent list of indices directly from the GitHub
        repository and not from the local copy.
    **kwargs : Keyword arguments
        Keyword arguments that can be passed to :code:`seaborn.heatmap()`.

    Returns
    -------
    Matplotlib Axes
        Axes object with the index heatmap.

    Examples
    --------
    >>> import spyndex
    >>> spyndex.plot.heatmap("NDVI","N","R",cmap = "Spectral",annot = True)

    If the index uses more than one band, the remaining bands or parameters must
    be passed in the :code:`params` dictionary.

    >>> spyndex.plot.heatmap(
    ...    "SAVI",
    ...    "N",
    ...    "R",
    ...    params = {"L": 0.5},
    ...    cmap = "Spectral",
    ...    annot = True)
    """

    grid = np.round(np.mgrid[0:1.1:0.1, 0:1.1:0.1], 1)
    X = grid[0].flatten()
    Y = grid[1].flatten()

    df = pd.DataFrame(
        {
            x: X,
            y: Y,
        }
    )

    if params is not None:
        params = {**params, x: df[x], y: df[y]}
    else:
        params = {x: df[x], y: df[y]}

    indices = _get_indices(online)
    _check_params(index, params, indices)

    df[index] = spyndex.computeIndex(index=index, params=params)

    df = df.pivot(index=y, columns=x, values=index)

    h = sns.heatmap(df, **kwargs)
    h.invert_yaxis()

    return h



"""spyndex - Awesome Spectral Indices in Python"""

__version__ = "0.10.0"
__author__ = "David Montero Loaiza <dml.mont@gmail.com>"
__all__ = []

from . import datasets, plot
from .axioms import bands, constants, indices
from .spyndex import *



import json
import os

import requests

import spyndex

from importlib.resources import files


def _load_JSON(file="spectral-indices-dict.json"):
    """Loads the specified JSON file from the data folder.

    Parameters
    ----------
    file : str
        File name.

    Returns
    -------
    object
        JSON file.
    """
    data_file = files("spyndex.data") / file
    with data_file.open("r", encoding="utf-8") as f:
        return json.load(f)


def _get_indices(online=False):
    """Retrieves the JSON of indices.

    Parameters
    ----------
    online : boolean
        Wheter to retrieve the most recent list of indices directly from the GitHub
        repository and not from the local copy.

    Returns
    -------
    dict
        Indices.
    """
    if online:
        indices = requests.get(
            "https://raw.githubusercontent.com/awesome-spectral-indices/awesome-spectral-indices/main/output/spectral-indices-dict.json"
        ).json()
    else:
        indices = _load_JSON()

    return indices["SpectralIndices"]


def _check_params(index: str, params: dict, indices: dict):
    """Checks if the parameters dictionary contains all required bands for the index
    computation.

    Parameters
    ----------
    index : str
        Index to check.
    params : dict
        Parameters dictionary to check.
    indices : dict
        Indices dictionary to check.

    Returns
    -------
    None
    """
    for band in indices[index]["bands"]:
        if band not in list(params.keys()):
            raise Exception(
                f"'{band}' is missing in the parameters for {index} computation!"
            )


def _has_ee_image(params: dict):
    """Checks if earthengine is installed and if params has an ee.Image.

    Parameters
    ----------
    params : dict
        Dictionary to check.

    Returns
    -------
    None
    """
    try:
        import ee
        return any(isinstance(v, ee.Image) for v in params.values())
    except ImportError:
        return False


def _has_ee_number(params: dict):
    """Checks if earthengine is installed and if params has an ee.Number.

    Parameters
    ----------
    params : dict
        Dictionary to check.

    Returns
    -------
    None
    """
    try:
        import ee
        return any(isinstance(v, ee.Number) for v in params.values())
    except ImportError:
        return False


def _maybe_import_earthengine(params: dict):
    """Import Earth Engine and eemont if any param is an Earth Engine object, or raise if missing.

    Parameters
    ----------
    params : dict
        Dictionary to check.

    Returns
    -------
    None
    """
    try:
        import ee
    except ImportError:
        needs_ee = False
        for v in params.values():
            if v.__class__.__name__ in ("Image", "Number"):
                needs_ee = True
                break
        if needs_ee:
            raise ImportError(
                "Earth Engine features require the optional dependency 'spyndex[ee]'.\n"
                "Install it with:\n\n    pip install 'spyndex[ee]'"
            )
        else:
            return

    if any(isinstance(v, (ee.Image, ee.Number)) for v in params.values()):
        try:
            import eemont
        except ImportError:
            raise ImportError(
                "Earth Engine features also require 'eemont'.\n"
                "Install it with:\n\n    pip install 'spyndex[ee]'"
            )




from box import Box

from .spyndex import computeIndex
from .utils import _get_indices, _load_JSON


class SpectralIndices(Box):
    """Spectral Indices object.

    This object allows interaction with the complete list of Spectral Indices in the
    Awesome Spectral Indices list.

    See Also
    --------
    SpectralIndex : Spectral Index object.

    Examples
    --------
    >>> import spyndex
    >>> spyndex.indices
    SpectralIndices(['AFRI1600', 'AFRI2100', 'ARI', ..., 'kNDVI', 'kRVI', 'kVARI'])
    """

    def __repr__(self):
        """Machine readable output of the Spectral Indices object."""

        return f"SpectralIndices({list(self.keys())})"

    def __str__(self):
        """Human readable output of the Spectral Indices object."""

        return f"{list(self.keys())}"


class SpectralIndex(object):
    """Spectral Index object.

    This object allows interaction with specific Spectral Indices in the
    Awesome Spectral Indices list. Attributes of the Spectral Index can be accessed
    and the index itself can be computed.

    See Also
    --------
    SpectralIndices : Spectral Indices object.

    Examples
    --------
    >>> import spyndex
    >>> spyndex.indices.NIRv
    SpectralIndex(NIRv: Near-Infrared Reflectance of Vegetation)
        * Application Domain: vegetation
        * Bands/Parameters: ('N', 'R')
        * Formula: ((N-R)/(N+R))*N
        * Reference: https://doi.org/10.1126/sciadv.1602244
    >>> spyndex.indices.NIRv.contributor
    'https://github.com/davemlz'
    >>> spyndex.indices.NIRv.compute(N = 0.67,R = 0.12)
    0.4664556962025317
    """

    def __init__(self, index: dict):

        self.short_name = index["short_name"]
        """Short name of the Spectral Index."""

        self.long_name = index["long_name"]
        """Long name of the Spectral Index."""

        self.bands = index["bands"]
        """Required bands and parameters for the Spectral Index computation."""

        self.application_domain = index["application_domain"]
        """Appication domain of the Spectral Index. One of ['vegetation', 'burn', 'water', 'urban', 'kernel', 'radar']"""

        self.reference = index["reference"]
        """URL to the reference/DOI of the Spectral Index."""

        self.formula = index["formula"].replace(" ", "")
        """Formula (as expression) of the Spectral Index."""

        self.date_of_addition = index["date_of_addition"]
        """Date of addition of the Spectral Index to Awesome Spectral Indices."""

        self.contributor = index["contributor"]
        """Contributor of the Spectral Index to Awesome Spectral Indices."""

        self.platforms = index["platforms"]
        """Platforms with the required bands for the Spectral Index computation."""

    def __repr__(self):
        """Machine readable output of the Spectral Index."""

        result = f"""SpectralIndex({self.short_name}: {self.long_name})
        * Application Domain: {self.application_domain}
        * Bands/Parameters: {self.bands}
        * Formula: {self.formula}
        * Reference: {self.reference}
        """

        return result

    def __str__(self):
        """Human readable output of the Spectral Index."""

        result = f"""{self.short_name}: {self.long_name}
        * Application Domain: {self.application_domain}
        * Bands/Parameters: {self.bands}
        * Formula: {self.formula}
        * Reference: {self.reference}
        """

        return result

    def compute(self, params=None, **kwargs):
        """Computes a Spectral Index.

        Parameters
        ----------
        params: dict
            Parameters used as inputs for the computation. The input data must be
            compatible with Overloaded Operators. Some inputs' types supported are pandas
            series, numpy arrays, xarray objects and numeric objects. Earth Engine objects
            are also compatible when using eemont.
        kwargs:
            Parameters used as inputs for the computation as keyword pairs. Ignored when
            params is defined.

        Returns
        -------
        Any
            Computed Spectral Index.

        Examples
        --------
        Compute a Spectral Index by passing the required :code:`params` dictionary:

        >>> import spyndex
        >>> spyndex.indices.NDVI.compute(
        ...     params = {
        ...         "N": 0.643,
        ...         "R": 0.175
        ...     }
        ... )
        0.5721271393643031

        Compute a Spectral Index by passing the required :code:`params` as keyword pairs:

        >>> spyndex.indices.NDVI.compute(N = 0.643, R = 0.175, L = 0.5)
        0.5721271393643031
        """

        if params is None:
            parameters = kwargs
        else:
            parameters = params

        return computeIndex(self.short_name, parameters)


def _create_indices():
    """Creates the set of Spectral Indices locally available."""

    indices = _get_indices(False)
    indices_class = {}
    for key, value in indices.items():
        indices_class[key] = SpectralIndex(value)

    return SpectralIndices(indices_class, frozen_box=True)


indices = _create_indices()


class Bands(Box):
    """Bands object.

    This object allows interaction with the list of bands required for the Spectral
    Indices in the Awesome Spectral Indices list.

    See Also
    --------
    Band : Band object.
    PlatformBand : PlatformBand object.

    Examples
    --------
    >>> import spyndex
    >>> spyndex.bands
    Constants(['L', 'g', 'C1', ..., 'sigma', 'p', 'c'])
    """

    def __repr__(self):
        """Machine readable output of the Constant."""

        return f"Bands({list(self.keys())})"

    def __str__(self):
        """Human readable output of the Constant."""

        return f"{list(self.keys())}"


class PlatformBand(object):
    """Platform Band object.

    This object shows information about a specific band for a specific sensor or
    paltform.

    See Also
    --------
    Bands : Bands object.
    Band : Band object.

    Examples
    --------
    >>> import spyndex
    >>> spyndex.bands.B.sentinel2a
    PlatformBand(Platform: Sentinel-2A, Band: Blue)
        * Band: B2
        * Center Wavelength (nm): 492.4
        * Bandwidth (nm): 66.0
    >>> spyndex.bands.B.sentinel2a.wavelength
    492.4
    """

    def __init__(self, platform_band: dict):

        self.platform = platform_band["platform"]
        """Name of the Platform."""

        self.band = platform_band["band"]
        """Band number/name for the specific Platform."""

        self.name = platform_band["name"]
        """Description/Name of the Band for the specific Platform."""

        self.wavelength = platform_band["wavelength"]
        """Center wavelength of the Band (in nm) for the specific Platform."""

        self.bandwidth = platform_band["bandwidth"]
        """Bandwidth of the Band (in nm) for the specific Platform."""

    def __repr__(self):
        """Machine readable output of the Band."""

        result = f"""PlatformBand(Platform: {self.platform}, Band: {self.name})
        * Band: {self.band}
        * Center Wavelength (nm): {self.wavelength}
        * Bandwidth (nm): {self.bandwidth}
        """

        return result

    def __str__(self):
        """Human readable output of the Band."""

        result = f"""Platform: {self.platform}, Band: {self.name}
        * Band: {self.band}
        * Center Wavelength (nm): {self.wavelength}
        * Bandwidth (nm): {self.bandwidth}
        """

        return result


class Band(object):
    """Band object.

    This object allows interaction with specific bands in the the list of required bands
    for the Spectral Indices in the Awesome Spectral Indices list. Attributes of the
    Band can be accessed using this object.

    See Also
    --------
    Bands : Bands object.
    PlatformBand : PlatformBand object.

    Examples
    --------
    >>> import spyndex
    >>> spyndex.bands.B
    Band(B: Blue)
    >>> spyndex.bands.B.long_name
    'Blue'
    """

    def __init__(self, band: dict):

        self.short_name = band["short_name"]
        """Short name of the Band."""

        self.long_name = band["long_name"]
        """Description/Name of the Band."""

        self.common_name = band["common_name"]
        """Common name of the Band according to the Electro-Optical Extension Specification for STAC."""

        self.min_wavelength = band["min_wavelength"]
        """Minimum wavelength of the spectral range of the band (nm)."""

        self.max_wavelength = band["max_wavelength"]
        """Maximum wavelength of the spectral range of the band (nm)."""

        self.standard = band["short_name"]
        """Short name of the Band. Equivalent to :code:`short_name`."""

        if "sentinel2a" in band["platforms"].keys():
            self.sentinel2a = PlatformBand(band["platforms"]["sentinel2a"])
            """Description of the band for the Sentinel-2A platform."""

        if "sentinel2b" in band["platforms"].keys():
            self.sentinel2b = PlatformBand(band["platforms"]["sentinel2b"])
            """Description of the band for the Sentinel-2B platform."""

        if "landsat4" in band["platforms"].keys():
            self.landsat4 = PlatformBand(band["platforms"]["landsat4"])
            """Description of the band for the Landsat 4 platform."""

        if "landsat5" in band["platforms"].keys():
            self.landsat5 = PlatformBand(band["platforms"]["landsat5"])
            """Description of the band for the Landsat 5 platform."""

        if "landsat7" in band["platforms"].keys():
            self.landsat7 = PlatformBand(band["platforms"]["landsat7"])
            """Description of the band for the Landsat 7 platform."""

        if "landsat8" in band["platforms"].keys():
            self.landsat8 = PlatformBand(band["platforms"]["landsat8"])
            """Description of the band for the Landsat 8 platform."""

        if "landsat9" in band["platforms"].keys():
            self.landsat9 = PlatformBand(band["platforms"]["landsat9"])
            """Description of the band for the Landsat 9 platform."""

        if "modis" in band["platforms"].keys():
            self.modis = PlatformBand(band["platforms"]["modis"])
            """Description of the band for the MODIS platform."""

        if "wv3" in band["platforms"].keys():
            self.worldview3 = PlatformBand(band["platforms"]["wv3"])
            """Description of the band for the WorldView-3 platform."""

        if "wv2" in band["platforms"].keys():
            self.worldview2 = PlatformBand(band["platforms"]["wv2"])
            """Description of the band for the WorldView-2 platform."""

        if "planetscope" in band["platforms"].keys():
            self.planetscope = PlatformBand(band["platforms"]["planetscope"])
            """Description of the band for the PlanetScope platform."""

    def __repr__(self):
        """Machine readable output of the Band."""

        result = f"""Band({self.short_name}: {self.long_name})
        """

        return result

    def __str__(self):
        """Human readable output of the Constant."""

        result = f"""{self.short_name}: {self.long_name}
        """

        return result


def _create_bands():
    """Creates the set of Bands locally available."""

    bands = _load_JSON("bands.json")
    bands_class = {}
    for key, value in bands.items():
        bands_class[key] = Band(value)

    return Bands(bands_class, frozen_box=True)


bands = _create_bands()


class Constants(Box):
    """Constants object.

    This object allows interaction with the list of constants of the Spectral Indices in
    the Awesome Spectral Indices list.

    See Also
    --------
    Constant : Constant object.
    Bands : Bands object.
    SpectralIndex: Spectral Index object.
    SpectralIndices : Spectral Indices object.

    Examples
    --------
    >>> import spyndex
    >>> spyndex.constants
    Constants(['L', 'g', 'C1', ..., 'sigma', 'p', 'c'])
    """

    def __repr__(self):
        """Machine readable output of the Constants object."""

        return f"Constants({list(self.keys())})"

    def __str__(self):
        """Human readable output of the Constants object."""

        return f"{list(self.keys())}"


class Constant(object):
    """Constant object.

    This object allows interaction with specific constants in the the list of constants of
    the Spectral Indices in  the Awesome Spectral Indices list. Attributes of the
    Constant can be accessed using this object.

    See Also
    --------
    Constants : Constants object.
    Bands : Bands object.
    SpectralIndex: Spectral Index object.
    SpectralIndices : Spectral Indices object.

    Examples
    --------
    >>> import spyndex
    >>> spyndex.constants.L
    Constant(L: Canopy background adjustment)
        * Default value: 1.0
    >>> spyndex.constants.L.default
    1.0
    """

    def __init__(self, constant: dict):

        self.description = constant["description"]
        """Description/Name of the Constant."""

        self.long_name = constant["description"]
        """Description/Name of the Constant. Equivalent to :code:`description`."""

        self.short_name = constant["short_name"]
        """Short name of the Constant."""

        self.standard = constant["short_name"]
        """Short name of the Constant. Equivalent to :code:`short_name`."""

        self.default = constant["default"]
        """Default value of the Constant."""

        self.value = constant["default"]
        """Default value of the Constant. Equivalent to :code:`default`."""

    def __repr__(self):
        """Machine readable output of the Constant."""

        result = f"""Constant({self.short_name}: {self.long_name})
        * Default value: {self.default}
        """

        return result

    def __str__(self):
        """Human readable output of the Constant."""

        result = f"""{self.short_name}: {self.long_name}
        * Default value: {self.default}
        """

        return result


def _create_constants():
    """Creates the set of Constants locally available."""

    constants = _load_JSON("constants.json")
    constants_class = {}
    for key, value in constants.items():
        constants_class[key] = Constant(value)

    return Constants(constants_class, frozen_box=True)


constants = _create_constants()



import unittest

import numpy as np
import pandas as pd
import xarray as xr

import spyndex

try:
    import ee
    import eemont
    ee.Initialize()
    HAS_EE = True
except ImportError:
    HAS_EE = False

B = np.random.normal(0.1, 0.1, 20 * 20)
G = np.random.normal(0.3, 0.1, 20 * 20)
R = np.random.normal(0.1, 0.1, 20 * 20)
N = np.random.normal(0.6, 0.1, 20 * 20)

df = pd.DataFrame({"B": B, "G": G, "R": R, "N": N})

da = xr.DataArray(
    np.array(
        [
            B.reshape(20, 20),
            G.reshape(20, 20),
            R.reshape(20, 20),
            N.reshape(20, 20),
        ]
    ),
    dims=("channel", "x", "y"),
    coords={"channel": ["B", "G", "R", "N"]},
)

indices = ["NDVI", "GNDVI", "SAVI", "EVI"]


class Test(unittest.TestCase):
    """Tests for the spyndex package."""

    def test_catalogue_indices(self):
        """Test the indices class"""
        self.assertIsInstance(spyndex.indices.NDVI.platforms, list)
        self.assertIsInstance(spyndex.indices.NDVI.application_domain, str)

    def test_catalogue_bands(self):
        """Test the bands class"""
        self.assertIsInstance(spyndex.bands.N.short_name, str)
        self.assertIsInstance(spyndex.bands.N.sentinel2a.wavelength, float)

    def test_catalogue_constants(self):
        """Test the constants class"""
        self.assertIsInstance(spyndex.constants.C1.short_name, str)

    def test_numeric(self):
        """Test the computeIndex() method"""
        result = spyndex.computeIndex(
            indices,
            {
                "N": 0.6,
                "R": 0.1,
                "G": 0.3,
                "B": 0.1,
                "L": spyndex.constants.L.default,
                "C1": spyndex.constants.C1.default,
                "C2": spyndex.constants.C2.default,
                "g": spyndex.constants.g.default,
            },
        )
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], float)

    def test_numeric_class(self):
        """Test the computeIndex() method"""
        result = spyndex.indices.NDVI.compute(
            {
                "N": 0.6,
                "R": 0.1,
            },
        )
        self.assertIsInstance(result, float)

    def test_numeric_class_kwargs(self):
        """Test the computeIndex() method"""
        result = spyndex.indices.NDVI.compute(
            N=0.6,
            R=0.1,
        )
        self.assertIsInstance(result, float)

    def test_numeric_kwargs(self):
        """Test the computeIndex() method"""
        result = spyndex.computeIndex(
            indices,
            N=0.6,
            R=0.1,
            G=0.3,
            B=0.1,
            L=spyndex.constants.L.default,
            C1=spyndex.constants.C1.default,
            C2=spyndex.constants.C2.default,
            g=spyndex.constants.g.default,
        )
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], float)

    def test_numeric_online(self):
        """Test the computeIndex() method"""
        result = spyndex.computeIndex(
            indices,
            {
                "N": 0.6,
                "R": 0.1,
                "G": 0.3,
                "B": 0.1,
                "L": spyndex.constants.L.default,
                "C1": spyndex.constants.C1.default,
                "C2": spyndex.constants.C2.default,
                "g": spyndex.constants.g.default,
            },
            online=True,
        )
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], float)

    def test_numpy(self):
        """Test the computeIndex() method"""
        result = spyndex.computeIndex(
            indices,
            {
                "N": N,
                "R": R,
                "G": G,
                "B": B,
                "L": spyndex.constants.L.default,
                "C1": spyndex.constants.C1.default,
                "C2": spyndex.constants.C2.default,
                "g": spyndex.constants.g.default,
            },
        )
        self.assertIsInstance(result, np.ndarray)

    def test_numpy_origin_false(self):
        """Test the computeIndex() method"""
        result = spyndex.computeIndex(
            indices,
            {
                "N": N,
                "R": R,
                "G": G,
                "B": B,
                "L": spyndex.constants.L.default,
                "C1": spyndex.constants.C1.default,
                "C2": spyndex.constants.C2.default,
                "g": spyndex.constants.g.default,
            },
            returnOrigin=False,
        )
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], np.ndarray)

    def test_pandas(self):
        """Test the computeIndex() method"""
        result = spyndex.computeIndex(
            indices,
            {
                "N": df["N"],
                "R": df["R"],
                "G": df["G"],
                "B": df["B"],
                "L": spyndex.constants.L.default,
                "C1": spyndex.constants.C1.default,
                "C2": spyndex.constants.C2.default,
                "g": spyndex.constants.g.default,
            },
        )
        self.assertIsInstance(result, pd.DataFrame)

    def test_pandas_origin_false(self):
        """Test the computeIndex() method"""
        result = spyndex.computeIndex(
            indices,
            {
                "N": df["N"],
                "R": df["R"],
                "G": df["G"],
                "B": df["B"],
                "L": spyndex.constants.L.default,
                "C1": spyndex.constants.C1.default,
                "C2": spyndex.constants.C2.default,
                "g": spyndex.constants.g.default,
            },
            returnOrigin=False,
        )
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], pd.Series)

    def test_xarray(self):
        """Test the computeIndex() method"""
        result = spyndex.computeIndex(
            indices,
            {
                "N": da.sel(channel="N"),
                "R": da.sel(channel="R"),
                "G": da.sel(channel="G"),
                "B": da.sel(channel="B"),
                "L": spyndex.constants.L.default,
                "C1": spyndex.constants.C1.default,
                "C2": spyndex.constants.C2.default,
                "g": spyndex.constants.g.default,
            },
        )
        self.assertIsInstance(result, xr.DataArray)

    def test_xarray_origin_false(self):
        """Test the computeIndex() method"""
        result = spyndex.computeIndex(
            indices,
            {
                "N": da.sel(channel="N"),
                "R": da.sel(channel="R"),
                "G": da.sel(channel="G"),
                "B": da.sel(channel="B"),
                "L": spyndex.constants.L.default,
                "C1": spyndex.constants.C1.default,
                "C2": spyndex.constants.C2.default,
                "g": spyndex.constants.g.default,
            },
            returnOrigin=False,
        )
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], xr.DataArray)

    @unittest.skipUnless(HAS_EE, "Earth Engine not installed")
    def test_ee(self):
        """Test the computeIndex() method"""
        result = spyndex.computeIndex(
            indices,
            {
                "N": ee.Image(0.63),
                "R": ee.Image(0.13),
                "G": ee.Image(0.32),
                "B": ee.Image(0.12),
                "L": spyndex.constants.L.default,
                "C1": spyndex.constants.C1.default,
                "C2": spyndex.constants.C2.default,
                "g": spyndex.constants.g.default,
            },
        )
        self.assertIsInstance(result, ee.Image)

    @unittest.skipUnless(HAS_EE, "Earth Engine not installed")
    def test_ee_origin_false(self):
        """Test the computeIndex() method"""
        result = spyndex.computeIndex(
            indices,
            {
                "N": ee.Image(0.63),
                "R": ee.Image(0.13),
                "G": ee.Image(0.32),
                "B": ee.Image(0.12),
                "L": spyndex.constants.L.default,
                "C1": spyndex.constants.C1.default,
                "C2": spyndex.constants.C2.default,
                "g": spyndex.constants.g.default,
            },
            returnOrigin=False,
        )
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], ee.Image)


if __name__ == "__main__":
    unittest.main()
